###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api import utils
from everysk.api.utils import EveryskList
from everysk.api.api_resources.api_resource import (
    APIResource,
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    CreateableAPIResource,
    FilterableAPIResource,
    UpdateableAPIResource
)
from everysk.core.unittests import TestCase, mock


###############################################################################
#   Mocking Classes Implementation
###############################################################################
class MockAPIResource(APIResource):
    @classmethod
    def class_name(cls):
        return 'testresource'

class MockAPIResourceEndsWithY(APIResource):
    @classmethod
    def class_name(cls):
        return 'testresourcey'

class MockAPIResourceEndsWithX(RetrievableAPIResource):
    @classmethod
    def class_name(cls):
        return 'testresourcex'

###############################################################################
#   APIResource TestCase Implementation
###############################################################################
@mock.patch.object(utils, 'create_api_requestor')
class APIResourceTestCase(TestCase):

    def setUp(self):
        self.retrieve_params = {'key1': 'value1', 'key2': 'value2'}
        self.params = {'param1': 'value1', 'param2': 'value2'}
        self.obj = APIResource(self.retrieve_params, self.params)

    def test_refresh_method(self, mock_utils_create_api: mock.MagicMock):
        resource = MockAPIResource({'param': 'value'}, {})
        resource.update = mock.MagicMock()
        resource.clear_unsaved_values = mock.MagicMock()
        api_requestor = mock_utils_create_api.return_value
        api_requestor.get.return_value = {'testresource': {'id': '123', 'name': 'Test Resource'}}
        result = resource.refresh()

        self.assertIsInstance(result, MockAPIResource)

    def test_class_name_list(self,  mock_utils_create_api: mock.MagicMock): # pylint: disable=unused-argument
        self.assertEqual(MockAPIResource.class_name_list(), 'testresources')

    def test_class_name_list_endswith_y(self,  mock_utils_create_api: mock.MagicMock): # pylint: disable=unused-argument
        self.assertEqual(MockAPIResourceEndsWithY.class_name_list(), 'testresourceies')

    def test_class_name_list_endswith_x(self,  mock_utils_create_api: mock.MagicMock): # pylint: disable=unused-argument
        self.assertEqual(MockAPIResourceEndsWithX.class_name_list(), 'testresourcexes')

    def test_class_url(self,  mock_utils_create_api: mock.MagicMock): # pylint: disable=unused-argument
        self.assertEqual(MockAPIResource.class_url(), '/testresources')

###############################################################################
#   Retrievable APIResource TestCase Implementation
###############################################################################
class MockRetrievableAPIResourceClassName(RetrievableAPIResource):

    @classmethod
    def class_name(cls):
        return 'retrievable'

class RetrievableAPIResourceTestCase(TestCase):

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_retrieve_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor

        mock_response_data = {'retrievable': {'id': '123', 'attribute': 'value'}}
        mock_api_requestor.get.return_value = mock_response_data
        test_params = {'key1': 'value1'}
        test_id = '123'
        expected_url = '/retrievables/123'
        response = MockRetrievableAPIResourceClassName.retrieve(test_id, **test_params)

        mock_api_requestor.get.assert_called_with(expected_url, test_params)
        self.assertIsInstance(response, MockRetrievableAPIResourceClassName)

###############################################################################
#   Listable APIResource TestCase Implementation
###############################################################################
class MockListableAPIResourceClassName(ListableAPIResource):

    @classmethod
    def class_name(cls):
        return 'listable'

class ListableAPIResourceTestCase(TestCase):

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_list_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor

        key_expected = MockListableAPIResourceClassName.class_name_list()
        mock_response_data = {
            key_expected: [{'id': '123', 'name': 'test_object'}]
        }
        mock_api_requestor.get.return_value = mock_response_data
        test_params = {'param1': 'value1'}
        result = MockListableAPIResourceClassName.list(**test_params)
        expected_url = MockListableAPIResourceClassName.class_url()

        mock_api_requestor.get.assert_called_once_with(expected_url, test_params)
        self.assertIsInstance(result, EveryskList)
        self.assertEqual(result[0]['id'], '123')
        self.assertEqual(result[0]['name'], 'test_object')

    @mock.patch('everysk.api.api_resources.api_resource.ListableAPIResource.list')
    def test_auto_paging_iter(self, mock_list):
        MockListableAPIResourceClassName = mock.MagicMock()
        MockListableAPIResourceClassName.class_url.return_value = '/test_url'
        MockListableAPIResourceClassName.class_name.return_value = 'items'

        first_page_response = {
            'items': [{'id': 3}, {'id': 4}],
            'next_page_token': 'token123'
        }
        second_page_response = {
            'items': [{'id': 3}, {'id': 4}],
            'next_page_token': None
        }
        first_page = EveryskList({}, first_page_response, 'items', MockListableAPIResourceClassName)
        second_page = EveryskList({'page_token': 'token123'}, second_page_response, 'items', MockListableAPIResourceClassName)
        mock_list.side_effect = [first_page, second_page]
        all_items = list(ListableAPIResource.auto_paging_iter())

        self.assertEqual(len(all_items), 4)

###############################################################################
#   Deletable APIResource TestCase Implementation
###############################################################################
class MockDeletableAPIResourceClassName(DeletableAPIResource):

    @classmethod
    def class_name(cls):
        return 'ResourceName'

class DeletableAPIResourceTestCase(TestCase):

    def setUp(self):
        retrieve_params = {'param1': 'value1'}
        params = {'params': 'value'}
        self.resource = MockDeletableAPIResourceClassName(retrieve_params, params)
        self.resource.id = '123'
        self.resource.workspace = 'default'

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_delete_success(self, mock_create_api_requestor):
        mock_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_requestor
        mock_response = {'ResourceName': {'id': '123', 'status': 'deleted'}}
        mock_requestor.delete.return_value = mock_response

        result = self.resource.delete()

        expected_url = '/ResourceNames/123?workspace=default'
        mock_requestor.delete.assert_called_with(expected_url)

        # Check the resulting state of the object
        self.assertEqual(result.id, '123')
        self.assertEqual(result.status, 'deleted')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_remove_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor
        mock_response = {'ResourceName': {'id': '123', 'attribute': 'value'}}
        mock_api_requestor.delete.return_value = mock_response

        # Call the class method 'remove'
        result = MockDeletableAPIResourceClassName.remove('123', workspace='default')

        # Verify API call
        expected_url = '/ResourceNames/123?workspace=default'
        mock_api_requestor.delete.assert_called_with(expected_url)

        # Verify the returned object
        self.assertIsInstance(result, MockDeletableAPIResourceClassName)
        self.assertEqual(result.id, '123')
        self.assertEqual(result.attribute, 'value')

###############################################################################
#   Creatable APIResource TestCase Implementation
###############################################################################
class MockCreatableAPIResourceClassName(CreateableAPIResource):

    @classmethod
    def class_name(cls):
        return 'creatable'

class CreatableAPIResourceTestCase(TestCase):

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_create_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor

        mock_response_data = {'creatable': {'key': 'data'}}
        mock_api_requestor.post.return_value = mock_response_data
        kwargs = {'param1': 'value1', 'param2': 'value2'}
        expected_url = '/creatables'
        result = MockCreatableAPIResourceClassName.create(**kwargs)

        mock_api_requestor.post.assert_called_once_with(expected_url, kwargs)
        self.assertIsInstance(result, MockCreatableAPIResourceClassName)

###############################################################################
#   Filterable APIResource TestCase Implementation
###############################################################################
class MockFilterableAPIResourceClassName(FilterableAPIResource):

    @classmethod
    def class_name(cls):
        return 'filterable'

class FilterableAPIResourceTestCase(TestCase):

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_filter_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_api_requestor.post.return_value = mock_api_requestor
        mock_create_api_requestor.return_value = mock_api_requestor

        key_expected = MockFilterableAPIResourceClassName.class_name_list()
        mock_response_data = {
            key_expected : [{'id': '123', 'name': 'test_object'}]
        }
        mock_api_requestor.post.return_value = mock_response_data
        test_params = {'param1': 'value1'}
        expected_url = f'{MockFilterableAPIResourceClassName.class_url()}/filter'
        result = MockFilterableAPIResourceClassName.filter(**test_params)

        mock_api_requestor.post.assert_called_once_with(expected_url, test_params)
        self.assertIsInstance(result, EveryskList)

###############################################################################
#   Updatable APIResource TestCase Implementation
###############################################################################
class MockUpdatableAPIResourceClassName(UpdateableAPIResource):
    @classmethod
    def class_name(cls):
        return 'updatable'

class UpdatableAPIResourceTestCase(TestCase):

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_modify_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_api_requestor.put.return_value = mock_api_requestor
        mock_create_api_requestor.return_value = mock_api_requestor

        mock_response = {
            'updatable': {'id': '123', 'attribute': 'value'}
        }
        mock_api_requestor.put.return_value = mock_response
        result = MockUpdatableAPIResourceClassName.modify('123', attr='new_value')

        mock_api_requestor.put.assert_called_once_with('/updatables/123', {'attr': 'new_value'})
        self.assertIsInstance(result, MockUpdatableAPIResourceClassName)

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_save_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor

        mock_response = {
            'updatable': {'id': '123', 'attribute': 'new_value'}
        }
        mock_api_requestor.put.return_value = mock_response
        retrieve_params = {'key1': 'value1', 'key2': 'value2'}
        params = {'param1': 'value1', 'param2': 'value2'}
        instance = MockUpdatableAPIResourceClassName(retrieve_params, params)
        instance.id = '123'
        unsaved_values = {'id': '123'}
        instance.save()
        expected_url = '/updatables/123'

        mock_api_requestor.put.assert_called_once_with(expected_url, unsaved_values)
        instance.update(mock_response['updatable'])
        self.assertEqual(instance.get('attribute'), 'new_value')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import os
from everysk.core.object import BaseObject
from everysk.core.cloud_function import main
from everysk.core.unittests import TestCase, mock


class CloudFunctionTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.old_redis_host = os.environ.get('REDIS_HOST', None)
        cls.old_redis_port = os.environ.get('REDIS_PORT', None)
        os.environ['REDIS_HOST'] = '127.0.0.1'
        os.environ['REDIS_PORT'] = '6379'
        cls.client = main.RedisClient()
        cls.redis_key = 'cloud-function-unit-test-redis-key'

    def setUp(self) -> None:
        self.client.connection.set(self.redis_key, 1, 1)
        self.context = BaseObject(resource='/project/collection/document')
        self.document = {
            'createTime': '2023-01-01T00:00:00+00:00',
            'fields': {
                'redis_key': {'stringValue': self.redis_key}
            },
            'name': '/project/collection/document',
            'updateTime': '2023-01-01T00:00:00+00:00'
        }

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.old_redis_host is not None:
            os.environ['REDIS_HOST'] = cls.old_redis_host
        else:
            del os.environ['REDIS_HOST']

        if cls.old_redis_port is not None:
            os.environ['REDIS_PORT'] = cls.old_redis_port
        else:
            del os.environ['REDIS_PORT']

    def test_redis_connection(self):
        client = main.RedisClient()
        self.assertIsInstance(client.connection, main.Redis)

    def test_redis_delete(self):
        self.assertEqual(self.client.connection.get(self.redis_key), b'1')
        self.client.delete(self.redis_key)
        self.assertIsNone(self.client.connection.get(self.redis_key))

    @mock.patch.object(main.log, 'info')
    def test_firestore_create(self, info: mock.MagicMock):
        event = {'oldValue': {}, 'updateMask': {}, 'value': self.document}
        main.firestore_cached_document_write(event, self.context)
        info.assert_called_once_with('Redis host: %s - %s', os.environ['REDIS_HOST'], self.redis_key)
        self.assertIsNone(self.client.connection.get(self.redis_key))

    @mock.patch.object(main.log, 'info')
    def test_firestore_delete(self, info: mock.MagicMock):
        event = {'oldValue': self.document, 'updateMask': {}, 'value': {}}
        main.firestore_cached_document_write(event, self.context)
        info.assert_called_once_with('Redis host: %s - %s', os.environ['REDIS_HOST'], self.redis_key)
        self.assertIsNone(self.client.connection.get(self.redis_key))

    @mock.patch.object(main.log, 'info')
    def test_firestore_update(self, info: mock.MagicMock):
        event = {'oldValue': self.document, 'updateMask': {}, 'value': self.document}
        main.firestore_cached_document_write(event, self.context)
        info.assert_called_once_with('Redis host: %s - %s', os.environ['REDIS_HOST'], self.redis_key)
        self.assertIsNone(self.client.connection.get(self.redis_key))

    @mock.patch.object(main.log, 'info')
    def test_firestore_other(self, info: mock.MagicMock):
        event = {'oldValue': {}, 'updateMask': {}, 'value': {}}
        main.firestore_cached_document_write(event, self.context)
        info.assert_not_called()
        self.assertEqual(self.client.connection.get(self.redis_key), b'1')



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from unittest import IsolatedAsyncioTestCase
from starlette.testclient import TestClient

from everysk.core.unittests import TestCase
from everysk.server.applications import create_application
from everysk.server.endpoints import BaseEndpoint
from everysk.server.middlewares import (
    BaseHTTPMiddleware,
    GZipMiddleware,
    Middleware,
    SecurityHeadersMiddleware,
    update_with_default_middlewares
)
from everysk.server.responses import Response
from everysk.server.routing import RouteLazy


class FakeMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request, call_next):
        return await call_next(request)


class UpdateMiddlewaresTestCase(TestCase):

    def test_update_with_default_middlewares_none(self):
        middlewares = update_with_default_middlewares(None)
        self.assertEqual(len(middlewares), 2)
        self.assertEqual(middlewares[0].cls, GZipMiddleware)
        self.assertEqual(middlewares[1].cls, SecurityHeadersMiddleware)

    def test_update_with_default_middlewares_empty(self):
        middlewares = update_with_default_middlewares([])
        self.assertEqual(len(middlewares), 2)
        self.assertEqual(middlewares[0].cls, GZipMiddleware)
        self.assertEqual(middlewares[1].cls, SecurityHeadersMiddleware)

    def test_update_with_default_middlewares_other(self):
        middlewares = update_with_default_middlewares([Middleware(FakeMiddleware)])
        self.assertEqual(len(middlewares), 3)
        self.assertEqual(middlewares[0].cls, GZipMiddleware)
        self.assertEqual(middlewares[1].cls, FakeMiddleware)
        self.assertEqual(middlewares[2].cls, SecurityHeadersMiddleware)


class FakeEndpoint(BaseEndpoint):

    async def get(self):
        return Response('Body', status_code=200)


class GZipMiddlewareTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', FakeEndpoint)])
        self.client = TestClient(self.app)

    async def test_gzip_middleware(self):
        response = self.client.get('/')
        self.assertEqual(response.headers['Content-Encoding'], 'gzip')
        self.assertEqual(response.content, b'Body')


class SecurityHeadersMiddlewareTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', FakeEndpoint)])
        self.client = TestClient(self.app)

    async def test_security_headers_middleware(self):
        response = self.client.get('/')
        self.assertEqual(response.headers['Strict-Transport-Security'], 'max-age=31536000; includeSubDomains')
        self.assertEqual(response.headers['X-Content-Type-Options'], 'nosniff')
        self.assertEqual(response.headers['X-DNS-Prefetch-Control'], 'off')
        self.assertEqual(response.headers['X-Download-Options'], 'noopen')
        self.assertEqual(response.headers['X-Frame-Options'], 'DENY')
        self.assertEqual(response.headers['X-Permitted-Cross-Domain-Policies'], 'none')



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access

from unittest import IsolatedAsyncioTestCase

import httpx
from starlette.testclient import TestClient

from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.exceptions import HttpError
from everysk.core.serialize import dumps
from everysk.core.unittests import TestCase, mock
from everysk.server.applications import create_application
from everysk.server.endpoints import BaseEndpoint, JSONEndpoint, HealthCheckEndpoint, RedirectEndpoint
from everysk.server.requests import Request, JSONRequest
from everysk.server.responses import Response
from everysk.server.routing import RouteLazy


class FakeEndpoint(BaseEndpoint):

    async def get(self):
        return Response(status_code=200)

    async def post(self):
        payload = await self.get_http_payload()
        if payload == b'error':
            raise HttpError(status_code=400, msg='Bad Request')
        if payload == b'error: log':
            raise HttpError(status_code=500, msg='Server error')

        return Response(payload)


class BaseEndpointTestCase(TestCase):

    def setUp(self) -> None:
        self.scope = {
            'type': 'http',
            'http_version': '1.1',
            'method': 'POST',
            'scheme': 'http',
            'path': '/',
            'headers': [
                    (b'content-type', b'application/json'),
                    (b'user-agent', b'Everysk Test Agent'),
                    (b'host', b'localhost:8000'),
                ],
            'client': ('127.0.0.1', 49192),
            'server': ('127.0.0.1', 8000),
        }

        self.endpoint = FakeEndpoint(scope=self.scope, receive=None, send=None)

    def test_base_endpoint(self):
        endpoint = FakeEndpoint(scope={'type': 'http'}, receive=None, send=None)
        self.assertIsInstance(endpoint.request, Request)
        self.assertEqual(endpoint._allowed_methods, ['GET', 'POST'])

    def test_base_endpoint_not_http(self):
        with self.assertRaises(HttpError) as context:
            FakeEndpoint(scope={'type': 'websocket'}, receive=None, send=None)
        self.assertEqual(context.exception.status_code, 500)
        self.assertEqual(context.exception.msg, 'Request is not an HTTP request.')

    def test_get_headers(self):
        self.assertDictEqual(
            self.endpoint.get_http_headers(),
            {'content-type': 'application/json', 'host': 'localhost:8000', 'user-agent': 'Everysk Test Agent'}
        )

    def test_get_http_method_function(self):
        self.assertEqual(self.endpoint.get_http_method_function(), self.endpoint.post)

    def test_get_http_method_name(self):
        self.assertEqual(self.endpoint.get_http_method_name(), 'post')

    def test_get_http_method_name_head(self):
        self.endpoint.scope['method'] = 'HEAD'
        self.assertEqual(self.endpoint.get_http_method_name(), 'get')


class BaseEndpointTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', FakeEndpoint)])
        self.client = TestClient(self.app)

    async def test_get_http_payload(self):
        response = self.client.post('/', data='payload')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'payload')

    async def test_get_http_exception_response(self):
        response = self.client.post('/', data='error')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'400 -> Bad Request')

    async def test_method_not_allowed(self):
        response = self.client.delete('/')
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.content, b'405 -> Method DELETE not allowed.')

    async def test_get_http_response(self):
        async def receive():
            return {'type': 'http.request', 'body': b''}

        endpoint = FakeEndpoint(scope={'type': 'http', 'method': 'POST'}, receive=receive, send=None)
        response = await endpoint.get_http_response()
        self.assertEqual(response.status_code, 200)

    async def test_dispatch(self):
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'')

    async def test_dispatch_error(self):
        response = self.client.post('/', data='error')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'400 -> Bad Request')

    @mock.patch('logging.Logger.log')
    async def test_dispatch_error_log(self, log: mock.MagicMock):
        response = self.client.post('/', data='error: log')
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b'500 -> Server error')
        log.assert_called_once_with(
            40,
            '500 -> Server error',
            extra={
                'http_payload': b'error: log',
                'labels': {},
                'traceback': (
                    'Traceback (most recent call last):\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 117, in dispatch\n'
                    '    response = await self.get_http_response()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 157, in get_http_response\n'
                    '    response = await method_function()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/_tests/endpoints.py", line 39, in post\n'
                    '    raise HttpError(status_code=500, msg=\'Server error\')\n'
                    'everysk.core.exceptions.HttpError: 500 -> Server error\n'
                ),
            },
            stacklevel=3
        )

    async def test_gzip_body(self):
        response = self.client.post('/', data=compress('Body', protocol='gzip', serialize=None), headers={'Content-Encoding': 'gzip'})
        self.assertEqual(response.content, b'Body')


class FakeJSONEndpoint(JSONEndpoint):

    async def get(self):
        return Response(status_code=200)

    async def post(self):
        payload = await self.get_http_payload()

        if payload == 'error':
            raise HttpError(status_code=400, msg='Bad Request')
        if payload == 'error: log':
            raise HttpError(status_code=500, msg='Server error')

        return {'payload': payload}

class FakeJSONEndpointPublic(FakeJSONEndpoint):
    rest_key_name: str = None
    rest_key_value: str = None

class FakeJSONEndpointPrivate(FakeJSONEndpoint):
    rest_key_name: str = 'key'
    rest_key_value: str = '12345'


class JSONEndpointTestCase(TestCase):

    def test_json_endpoint(self):
        endpoint = FakeJSONEndpoint(scope={'type': 'http'}, receive=None, send=None)
        self.assertIsInstance(endpoint.request, JSONRequest)
        self.assertEqual(endpoint._allowed_methods, ['GET', 'POST'])
        self.assertEqual(endpoint.rest_key_name, settings.EVERYSK_SERVER_REST_KEY_NAME)
        self.assertEqual(endpoint.rest_key_value, settings.EVERYSK_SERVER_REST_KEY_VALUE)

    def test_check_rest_key_default(self):
        endpoint = FakeJSONEndpoint(scope={'type': 'http', 'headers': [(b'x-rest-key', b'12345')]}, receive=None, send=None)
        self.assertTrue(endpoint.check_rest_key())

    def test_check_rest_key_none(self):
        endpoint = FakeJSONEndpoint(scope={'type': 'http', 'headers': []}, receive=None, send=None)
        self.assertFalse(endpoint.check_rest_key())

    def test_check_rest_key_wrong(self):
        endpoint = FakeJSONEndpoint(scope={'type': 'http', 'headers': [(b'x-rest-key', b'54321')]}, receive=None, send=None)
        self.assertFalse(endpoint.check_rest_key())

    def test_check_rest_key_public_default(self):
        endpoint = FakeJSONEndpointPublic(scope={'type': 'http', 'headers': [(b'x-rest-key', b'12345')]}, receive=None, send=None)
        self.assertTrue(endpoint.check_rest_key())

    def test_check_rest_key_public_none(self):
        endpoint = FakeJSONEndpointPublic(scope={'type': 'http', 'headers': []}, receive=None, send=None)
        self.assertTrue(endpoint.check_rest_key())

    def test_check_rest_key_public_wrong(self):
        endpoint = FakeJSONEndpointPublic(scope={'type': 'http', 'headers': [(b'x-rest-key', b'54321')]}, receive=None, send=None)
        self.assertTrue(endpoint.check_rest_key())

    def test_check_rest_key_private_default(self):
        endpoint = FakeJSONEndpointPrivate(scope={'type': 'http', 'headers': [(b'x-rest-key', b'12345')]}, receive=None, send=None)
        self.assertFalse(endpoint.check_rest_key())

    def test_check_rest_key_private_none(self):
        endpoint = FakeJSONEndpointPrivate(scope={'type': 'http', 'headers': []}, receive=None, send=None)
        self.assertFalse(endpoint.check_rest_key())

    def test_check_rest_key_private_wrong(self):
        endpoint = FakeJSONEndpointPrivate(scope={'type': 'http', 'headers': [(b'x-rest-key', b'54321')]}, receive=None, send=None)
        self.assertFalse(endpoint.check_rest_key())

    def test_check_rest_key_private_custom(self):
        endpoint = FakeJSONEndpointPrivate(scope={'type': 'http', 'headers': [(b'key', b'12345')]}, receive=None, send=None)
        self.assertTrue(endpoint.check_rest_key())


class JSONEndpointTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', FakeJSONEndpointPublic)])
        self.client = TestClient(self.app)

    async def test_request_response(self):
        data = dumps({'data': Undefined}, protocol='json', use_undefined=True)
        response = self.client.post('/', data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {'payload': {'data': {'__undefined__': None}}})

    async def test_get_http_payload(self):
        response = self.client.post('/', json='payload')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'{"payload":"payload"}')

    async def test_get_http_exception_response(self):
        response = self.client.post('/', json='error')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'{"error":"400 -> Bad Request","code":400,"trace_id":""}')

    async def test_method_not_allowed(self):
        response = self.client.delete('/')
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.content, b'{"error":"405 -> Method DELETE not allowed.","code":405,"trace_id":""}')

    async def test_get_http_response(self):
        async def receive():
            return {'type': 'http.request', 'body': b''}

        endpoint = FakeEndpoint(scope={'type': 'http', 'method': 'POST'}, receive=receive, send=None)
        response = await endpoint.get_http_response()
        self.assertEqual(response.status_code, 200)

    async def test_dispatch(self):
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'')

    async def test_dispatch_error(self):
        response = self.client.post('/', json='error')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'{"error":"400 -> Bad Request","code":400,"trace_id":""}')

    @mock.patch('logging.Logger.log')
    async def test_dispatch_error_log(self, log: mock.MagicMock):
        response = self.client.post('/', json='error: log')
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b'{"error":"500 -> Server error","code":500,"trace_id":""}')
        log.assert_called_once_with(
            40,
            '500 -> Server error',
            extra={
                'http_payload': 'error: log',
                'labels': {},
                'traceback': (
                    'Traceback (most recent call last):\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 117, in dispatch\n'
                    '    response = await self.get_http_response()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 255, in get_http_response\n'
                    '    response = await super().get_http_response()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 157, in get_http_response\n'
                    '    response = await method_function()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/_tests/endpoints.py", line 174, in post\n'
                    '    raise HttpError(status_code=500, msg=\'Server error\')\n'
                    'everysk.core.exceptions.HttpError: 500 -> Server error\n'
                ),
            },
            stacklevel=3
        )

    @mock.patch('logging.Logger.log')
    async def test_dispatch_error_log_trace(self, log: mock.MagicMock):
        response = self.client.post('/', json='error: log', headers={'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'})
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b'{"error":"500 -> Server error","code":500,"trace_id":"projects/None/traces/4bfa9e049143840bef864a7859f2e5df"}')
        log.assert_called_once_with(
            40,
            '500 -> Server error',
            extra={
                'http_headers': {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'},
                'http_payload': 'error: log',
                'labels': {},
                'traceback': (
                    'Traceback (most recent call last):\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 117, in dispatch\n'
                    '    response = await self.get_http_response()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 255, in get_http_response\n'
                    '    response = await super().get_http_response()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/endpoints.py", line 157, in get_http_response\n'
                    '    response = await method_function()\n'
                    '               ^^^^^^^^^^^^^^^^^^^^^^^\n'
                    '  File "/var/app/src/everysk/server/_tests/endpoints.py", line 174, in post\n'
                    '    raise HttpError(status_code=500, msg=\'Server error\')\n'
                    'everysk.core.exceptions.HttpError: 500 -> Server error\n'
                ),
            },
            stacklevel=3
        )

    async def test_gzip_body(self):
        response = self.client.post('/', data=compress('Body', protocol='gzip', serialize='json'), headers={'Content-Encoding': 'gzip'})
        self.assertEqual(response.content, b'{"payload":"Body"}')


class HealthCheckEndpointTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', HealthCheckEndpoint)])
        self.client = TestClient(self.app)

    async def test_health_check_get(self):
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'{"status":"SENTA_A_PUA"}')

    async def test_health_check_post(self):
        response = self.client.post('/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'{"status":"SENTA_A_PUA"}')


class FakeRedirectEndpoint(RedirectEndpoint):
    host_url: str = 'https://www.example.com'


class RedirectEndpointTestCase(TestCase):

    def test_required_attributes(self):
        with self.assertRaises(ValueError) as context:
            RedirectEndpoint(scope={'type': 'http'}, receive=None, send=None)
        self.assertEqual(str(context.exception), 'host_url is required for redirect endpoint.')

    def test_get_host(self):
        endpoint = FakeRedirectEndpoint(scope={'type': 'http'}, receive=None, send=None)
        self.assertEqual(endpoint.get_host(), 'www.example.com')

    def test_get_timeout(self):
        endpoint = FakeRedirectEndpoint(scope={'type': 'http'}, receive=None, send=None)
        self.assertEqual(endpoint.get_timeout(), httpx.Timeout(timeout=30, read=endpoint.timeout))


class RedirectEndpointTestCaseAsync(IsolatedAsyncioTestCase):
    # pylint: disable=unnecessary-dunder-call

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/path', FakeRedirectEndpoint)])
        self.client = TestClient(self.app)

    @mock.patch('httpx.Client')
    async def test_redirect_get_text(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.get.return_value = httpx.Response(text='Text', status_code=200)
        response = self.client.get('/path?qs=1')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/plain; charset=utf-8')
        self.assertEqual(response.content, b'"Text"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient'}),
            mock.call().__enter__(),
            mock.call().__enter__().get('https://www.example.com/path?qs=1', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_get_html(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.get.return_value = httpx.Response(html='<b>Text</b>', status_code=200)
        response = self.client.get('/path?qs=1')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/html; charset=utf-8')
        self.assertEqual(response.content, b'"<b>Text</b>"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient'}),
            mock.call().__enter__(),
            mock.call().__enter__().get('https://www.example.com/path?qs=1', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_get_json(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.get.return_value = httpx.Response(json={'key': 'Text'}, status_code=200)
        response = self.client.get('/path?qs=1')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'application/json')
        self.assertEqual(response.content, b'{"key": "Text"}')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient'}),
            mock.call().__enter__(),
            mock.call().__enter__().get('https://www.example.com/path?qs=1', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_data_text(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(text='Text', status_code=200)
        response = self.client.post('/path', data={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/plain; charset=utf-8')
        self.assertEqual(response.content, b'"Text"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '9', 'content-type': 'application/x-www-form-urlencoded'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'key=value', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_data_html(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(html='<b>Text</b>', status_code=200)
        response = self.client.post('/path', data={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/html; charset=utf-8')
        self.assertEqual(response.content, b'"<b>Text</b>"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '9', 'content-type': 'application/x-www-form-urlencoded'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'key=value', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_data_json(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(json={'key': 'Text'}, status_code=200)
        response = self.client.post('/path', data={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'application/json')
        self.assertEqual(response.content, b'{"key": "Text"}')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '9', 'content-type': 'application/x-www-form-urlencoded'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'key=value', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_json_text(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(text='Text', status_code=200)
        response = self.client.post('/path', json={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/plain; charset=utf-8')
        self.assertEqual(response.content, b'"Text"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '16', 'content-type': 'application/json'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'{"key": "value"}', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_json_html(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(html='<b>Text</b>', status_code=200)
        response = self.client.post('/path', json={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'text/html; charset=utf-8')
        self.assertEqual(response.content, b'"<b>Text</b>"')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '16', 'content-type': 'application/json'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'{"key": "value"}', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])

    @mock.patch('httpx.Client')
    async def test_redirect_post_json_json(self, client: mock.MagicMock):
        client.return_value.__enter__.return_value.post.return_value = httpx.Response(json={'key': 'Text'}, status_code=200)
        response = self.client.post('/path', json={'key': 'value'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get('content-type'), 'application/json')
        self.assertEqual(response.content, b'{"key": "Text"}')
        client.assert_has_calls([
            mock.call(headers={'host': 'www.example.com', 'accept': '*/*', 'accept-encoding': 'gzip, deflate', 'connection': 'keep-alive', 'user-agent': 'testclient', 'content-length': '16', 'content-type': 'application/json'}),
            mock.call().__enter__(),
            mock.call().__enter__().post('https://www.example.com/path', content=b'{"key": "value"}', timeout=httpx.Timeout(connect=30, read=600, write=30, pool=30)),
            mock.call().__exit__(None, None, None)
        ])



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import string
import secrets
import uuid

from everysk.config import settings

###############################################################################
#   Globals
###############################################################################
ALPHANUMERIC = string.ascii_letters + string.digits
SYMBOL_ID_MAX_LEN = settings.SYMBOL_ID_MAX_LEN

###############################################################################
#   Cryptography Implementation
###############################################################################
def generate_random_id(length: int, characters: str = ALPHANUMERIC) -> str:
    """
    Generate a random ID based on the specified length and characters.

    Args:
        length (int): The desired length for the random ID.
        characters (str, optional): The characters to use for generating the ID.
            Default is alphanumeric characters.

    Returns:
        str: The generated random ID consisting of the specified characters.

    Example:
        >>> generate_random_id(18)
        'xiXwQuig6cI11ny095'

        >>> generate_random_id(5, 'abc')
        'cbacb'
    """
    return ''.join(secrets.choice(characters) for i in range(length))

def generate_unique_id():
    """
    Generate a unique ID with fixed length (32 characters).

    Returns:
        str: The generated unique ID consisting of hexadecimal characters.

    Example:
        >>> generate_unique_id()
        'dbdc803042ff48e19a344ad080a102dd'
    """
    return uuid.uuid4().hex

def generate_short_random_id() -> str:
    """
    Generate a random ID with a fixed length of 8 characters.

    Returns:
        str: The generated unique ID consisting of alphanumeric characters.

    Example:
        >>> generate_short_random_id()
        'u0tmNqHP'
    """
    return generate_random_id(length=settings.SIMPLE_UNIQUE_ID_LENGTH)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from collections.abc import Iterable
from typing import Any
from everysk.core.exceptions import FieldValueError

###############################################################################
#   Entity List Implementation
###############################################################################
class EntityList(list):
    """
    A list subclass with validation capabilities.

    This class extends the standard Python list and adds validation for the values
    added or inserted into it.

    Attributes:
        min_size (int): The minimum allowed size for the list (default is None).
        max_size (int): The maximum allowed size for the list (default is None).

    Example:
        To create a custom list with validation capabilities:
        my_list = EntityList()
    """
    min_size: int | None = None
    max_size: int | None = None

    def __init__(self, *args) -> None:
        """
        Initialize the EntityList.

        Args:
            *args: Optional initial elements to populate the list.

        Example:
            To create an EntityList with initial elements:
            my_list = EntityList([1, 2, 3])
        """
        if args:
            try:
                args = ([self._validate(item) for item in args[0]], )
            except TypeError:
                raise FieldValueError('Unsupported format.') from TypeError
        super().__init__(*args)

    def __setitem__(self, index: int, value: Any) -> None:
        """
        Set an item in the list with validation.

        Args:
            index (int): The index at which to set the value.
            value: The value to be set.
        """
        value = self._validate(value)
        return super().__setitem__(index, value)

    def insert(self, __index: int, __object: Any) -> None:
        """
        Insert an object at a specified index with validation.

        Args:
            __index (int): The index at which to insert the object.
            __object: The object to be inserted.
        """
        __object = self._validate(__object)
        return super().insert(__index, __object)

    def append(self, __object: Any) -> None:
        """
        Append an object to the list with validation.

        Args:
            __object: The object to be appended.
        """
        __object = self._validate(__object)
        return super().append(__object)

    def extend(self, __iterable: Iterable) -> None:
        """
        Extend the list with values from an iterable with validation.

        Args:
            __iterable (Iterable): An iterable containing values to be added to the list.
        """
        for value in __iterable:
            value = self._validate(value)
        return super().extend(__iterable)

    def _validate(self, value: Any) -> Any:
        """
        Validate a value.

        Args:
            value: The value to be validated.

        Raises:
            NotImplementedError: This method should be overridden in subclasses.

        Returns:
            The validated value.
        """
        raise NotImplementedError()

    def to_native(self, add_class_path: str | None = None, recursion: bool = False) -> Any:
        """
        Converts the object to the specified Python type.

        Args:
            add_class_path (str | None, optional): The class path to add when converting the object. Defaults to None.
            recursion (bool, optional): Indicates whether to recursively convert nested objects. Defaults to False.

        Returns:
            object: The converted object.

        """
        return self.to_list(add_class_path=add_class_path, recursion=recursion)

    def to_list(self, add_class_path: bool = False, recursion: bool = False) -> list[dict]: # pylint: disable=unused-argument
        """
        Convert the Securities collection to a list of dictionaries.
        """
        return list(self)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Self, Any

from everysk.config import settings
from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import SDKValueError, FieldValueError
from everysk.core.fields import StrField, DateTimeField, ListField
from everysk.core.object import MetaClass as ObjMetaClass, BaseDictConfig
from everysk.core.string import is_string_object
from everysk.core.lists import split_in_slices

from everysk.sdk.base import BaseSDK, BaseDict
from everysk.sdk.engines.cryptography import generate_random_id
from everysk.sdk.entities import Query
from everysk.sdk.entities.script import Script
from everysk.sdk.entities.tags import Tags

###############################################################################
#   Query MetaClass Implementation
###############################################################################
class QueryMetaClass(ObjMetaClass):
    """
    Metaclass for the Query class that allows for the query attribute to be accessed
    directly from the entity class.

    Example:
        To access the query attribute from the entity class:
        >>> MyClass.query
        Query()
    """
    def __getattribute__(cls, __name: str) -> Any:
        """
        Get the query attribute from the entity class.
        """
        if __name == 'query':
            return Query(cls)
        return super().__getattribute__(__name)

###############################################################################
#   Script MetaClass Implementation
###############################################################################
class ScriptMetaClass(QueryMetaClass):
    """
    Metaclass for the Script class that allows for the script attribute to be accessed
    directly from the entity class.

    Example:
        To access the script attribute from the entity class:
        >>> MyClass.script
        Script()

    Notes:
        This metaclass overrides the __getatrribute__ method to enable direct access
    """
    def __getattribute__(cls, __name: str) -> Any:
        """
        Get the script attribute from the entity class.
        """
        if __name == 'script':
            return Script(cls)
        return super().__getattribute__(__name)

###############################################################################
#   BaseEntityConfig Implementation
###############################################################################
class BaseEntityConfig(BaseDictConfig):
    exclude_keys: frozenset[str] = frozenset(['query', 'script', '_is_frozen', '_silent', '_errors', '_orderable_attributes'])
    keys_blacklist: frozenset[str] = frozenset(['query', 'script'])

###############################################################################
#   Base Entity Implementation
###############################################################################
class BaseEntity(BaseSDK, BaseDict, metaclass=QueryMetaClass):
    """
    Base class for all entities in the SDK library that provides common functionality. This class
    should not be instantiated directly, but rather should be subclassed by other entity classes.

    Attributes:
        id (str): The unique identifier of the entity.
        version (str): The version of the entity.
        created_on (DateTime): The date and time the entity was created.
        updated_on (DateTime): The date and time the entity was last updated.

    Example:
        To create a new entity:
        >>> my_entity = MyEntity(id="my_id", workspace="my_workspace", name="my_name", description="my_description")
    """
    class Config(BaseEntityConfig):
        pass

    _config: Config = None
    _orderable_attributes = ListField(default=['id', 'created_on', 'updated_on'], readonly=True)

    id = StrField(default=None)

    query: Query = None

    created_on = DateTimeField(default=Undefined, empty_is_none=True, required_lazy=True)
    updated_on = DateTimeField(default=Undefined, empty_is_none=True, required_lazy=True)

    version = StrField(default=settings.ENTITY_DEFAULT_VERSION, required_lazy=True)

    def __after_init__(self) -> None:
        """
        Method that runs after the __init__ method.
        This method must return None.
        """
        super().__after_init__()
        if self.created_on is Undefined:
            self.created_on = DateTime.now()
        if self.updated_on is Undefined:
            self.updated_on = self.created_on

    def _process_date(self, value: DateTime | None) -> str | None:
        """
        Process the date value.

        Args:
            value (DateTime): The date value to process.

        Returns:
            str: The processed date value.
        """
        return Date.strftime_or_null(value)

    def _process_tags(self, value: Tags | None) -> list | None:
        """
        Convert the entity to a JSON-serializable dictionary.
        This method converts the entity object into a dictionary that can be easily
        serialized to JSON.

        Args:
            value (Tags): The tags value to process.

        Returns:
            list: The processed tags value.
        """
        return value.to_list() if isinstance(value, Tags) else value

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if add_class_path is False:
            if 'date' in self:
                dct['date_time'] = DateTime.strftime_or_null(self.date) # pylint: disable=no-member

            if 'created_on' in self:
                dct.pop('created_on')
                dct['created'] = self.created_on.timestamp() if self.created_on is not None else None

            if 'updated_on' in self:
                dct.pop('updated_on')
                dct['updated'] = self.updated_on.timestamp() if self.updated_on is not None else None

        dct.pop('query', None)
        dct.pop('script', None)

        return dct

    @staticmethod
    def get_id_prefix() -> str:
        """
        Get the prefix for the unique identifier for this entity.

        Returns:
            str: The prefix for the unique identifier.

        Raises:
            NotImplementedError: This method should be overridden in subclasses.
        """
        raise NotImplementedError()

    def generate_id(self) -> str:
        """
        Generate a unique ID for an entity instance.

        Returns:
            str: The generated unique ID.

        Example:
            To generate a unique ID for an entity instance:
            >>> unique_id = MyEntity().generate_id()
        """
        prefix: str = self.get_id_prefix()
        unique_id: str = generate_random_id(length=settings.ENTITY_ID_LENGTH)
        return f'{prefix}{unique_id}'

    @classmethod
    def validate_id(cls, entity_id: str) -> bool:
        """
        Validate an entity's ID.

        Args:
            entity_id str: The ID to be validated.

        Returns:
            bool: True if the ID is valid, False otherwise.

        Example:
            To validate an entity's ID:
            >>> is_valid = MyEntity.validate_id(my_id)
        """
        if entity_id:
            try:
                cls(id=entity_id)
                return True
            except Exception: # pylint: disable=broad-exception-caught
                pass
        return False

    def validate(self) -> bool:
        """
        Validate the entity's attributes.

        This method performs validation checks on the entity's attributes to ensure they meet
        the required criteria. If all required fields are present, the validation is considered
        successful and the method returns True. If any required field is missing, it raises a
        RequiredFieldError exception.

        Args:
            self (Self): The entity instance to validate.

        Returns:
            bool: True if the validation is successful.

        Raises:
            RequiredFieldError: If any required field is missing.

        Example:
            To validate an entity:

            >>> is_valid = my_entity.validate()
            >>> if is_valid:
            >>>     # Handle the valid entity
            >>> else:
            >>>     # Handle the invalid entity
        """
        self.validate_required_fields()
        return True

    def _pre_validate(self) -> None:
        self.id = self.generate_id() # pylint: disable=attribute-defined-outside-init

    def _pos_validate(self) -> None:
        self.id = None # pylint: disable=attribute-defined-outside-init
        self.created_on = None
        self.updated_on = None

    @classmethod
    def validate_transient(cls, entity_dict: dict) -> Self:
        """
        Validate the entity properties.

        Args:
            entity_dict (dict): The entity properties.

        Returns:
            BaseEntity: The entity object.

        Example:
            >>> entity_dict = {'name': 'My Entity'}
            >>> BaseEntity.validate_transient(entity_dict)
        """
        # Set the entity properties
        entity = cls(**entity_dict)
        entity._pre_validate()
        entity.validate()
        entity._pos_validate()

        return entity

    @classmethod
    def check(cls, entity_dict: dict) -> Self:
        """
        Check the entity properties.

        Args:
            entity_dict (dict): The entity properties.

        Returns:
            BaseEntity: The entity object.

        Raises:
            FieldValueError: If the entity properties are invalid.
            RequiredFieldError: If a required field is missing.
        """
        entity: Self = cls(**entity_dict)
        entity.validate()
        return entity

    def _check_query(self, query: Query) -> bool:
        """
        Check the query object.

        Args:
            query (Query): The query object.

        Returns:
            bool: True if the query object is valid.
        """
        # pylint: disable=no-member
        if self.name and 'date' in query.order:
            raise SDKValueError("Can't filter by Name and Date at the same time,  must order by updated_on")

        return True

    def _check_entity_to_query(self) -> bool:
        """
        Check the entity object to query.

        Returns:
            bool: True if the entity object is valid.
        """
        # pylint: disable=no-member
        if self.name and self.tags:
            raise SDKValueError("Can't filter by Name and Tags at the same time")
        if self.name and self.link_uid:
            raise SDKValueError("Can't filter by Name and Link UID at the same time")

        return True

    def _mount_query(self, query: Query) -> Query:
        """
        Mount the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        # pylint: disable=no-member
        if self.workspace is not None:
            query = query.where('workspace', self.workspace)
        if self.link_uid is not None:
            query = query.where('link_uid', self.link_uid)
        if self.name is not None:
            query = query.where('name', self.name)
        if self.date is not None:
            query = query.where('date', self.date)
        if self.tags:
            query = query.where('tags', self.tags)

        return query

    def to_query(self, order: list | None = None, projection: list | None = None, distinct_on: list | None = None,
                 limit: int | None = None, offset: int | None = None, page_size: int | None = None, page_token: str | None = None) -> Query:
        """
        This method converts the entity object into a query object.

        Args:
            order (List[str], optional): The order to apply to the query. Defaults to None.
            projection (List[str], optional): The projection to apply to the query. Defaults to None.
            distinct_on (List[str], optional): The distinct_on to apply to the query. Defaults to None.
            limit (int, optional): The limit to apply to the query. Defaults to None.
            offset (int, optional): The offset to apply to the query. Defaults to None.
            page_size (int, optional): The page size to apply to the query. Defaults to None.
            page_token (str, optional): The page token to apply to the query. Defaults to None.

        Returns:
            Query: A query object representing the entity.

        Example:
            To convert an entity object into a query object:
            >>> entity = MyClass(property1="value1", property2="value2")
            >>> query = entity.to_query(order=['property1'], limit=10)
        """
        self._check_entity_to_query()

        query: Query = Query(
            self.__class__,
            order=order,
            projection=projection,
            distinct_on=distinct_on,
            limit=limit,
            offset=offset,
            page_size=page_size,
            page_token=page_token
        )

        self._check_query(query)
        return self._mount_query(query)

    @classmethod
    def _normalize_projection(cls, projection: str | list[str]) -> list[str]:
        """
        Validate the projection attributes for the query and return the instance.

        This method validate the desired properties that should be returned in entity.
        The properties can either be set to include (using the property name) or to exclude
        (prefixing the property name with '-'). Both inclusion and exclusion should not be set
        in the same projection.

        Args:
            - projection (Union[List, str]): A property name as a string or a list of property names
            indicating which properties to include or exclude in the entity.

        Returns:
            - Query: The instance of the current object.

        Raises:
            - ValueError: If both projection and inverse projection are set in the same projection or
            if the projection properties do not belong to the entity kind.

        Example:
            To create a projection with a projection condition:
            >>> projection = self._validate_projection('property_name')

            To create a query with an inverse projection condition:
            >>> projection = self._validate_projection('-property_name')

            To create a query with a projection condition using a list:
            >>> projection = self._validate_projection(['property_name_1', 'property_name_2'])
        """
        if projection is None:
            return []

        if is_string_object(projection):
            projection = [projection]

        count: int = sum(property_name.startswith('-') for property_name in projection)
        if not (count == 0 or count == len(projection)):
            raise ValueError('Projection and Inverse Projection should not be set in the same query')

        entity_properties: set = set(cls.__attributes__.keys()) # pylint: disable=protected-access
        projection_properties: set = set([property_name.replace('-', '') for property_name in projection])
        difference: set = projection_properties.difference(entity_properties)

        if difference:
            difference_: str = ', '.join(difference)
            raise ValueError(f'Projection properties does not belongs to {cls.__name__}: {difference_}')

        return projection

    @classmethod
    def retrieve(cls, entity_id: str, projection: list | str | None = None) -> Self | None:
        """
        Retrieve an entity by its ID.

        Args:
            entity_id (str): The unique identifier of the entity to retrieve.
            projection (Union[List, str], optional): A property name as a string or a list of property names

        Returns:
            Self: An instance of the class representing the retrieved entity.
            None: If not found

        Example:
            To retrieve an entity by its ID:

            >>> entity = MyClass.retrieve("entity_id_here")
            >>> if entity:
            >>>    # Handle the retrieved entity
            >>> else:
            >>>    # Entity not found
        """
        if projection is not None:
            projection = cls._normalize_projection(projection)

        entity_dict: dict | None = cls.get_response(params={'entity_id': entity_id, 'projection': projection})

        if entity_dict is None:
            return None

        return cls(**entity_dict)

    @classmethod
    def create(cls, entity_dict: dict) -> Self:
        """
        Create a new entity using provided attributes from a dictionary.

        Args:
            entity_dict (dict): A dictionary representing the entity's attributes.

        Returns:
            Self: An instance of the class representing the newly created entity.

        Example:
            To create a new entity with attributes from a dictionary and optional keyword arguments:

            >>> entity_data = {'property1': value1, 'property2': value2}
            >>> new_entity = MyClass.create(entity_data)
        """
        entity_dict_: dict = cls.get_response(params={'entity_dict': entity_dict})

        return cls(**entity_dict_)

    @classmethod
    def modify(cls, entity_id: str, overwrites: dict) -> Self | None:
        """
        Modify an existing entity by updating its attributes using the provided overwrites.

        Args:
            entity_id (str): The unique identifier of the entity to modify.
            overwrites (dict): A dictionary containing attribute updates to apply to the entity.

        Returns:
            Self: An instance of the class representing the modified entity.
            None: If not found

        Example:
            To modify an existing entity by updating its attributes with overwrites:

            >>> entity_id_to_modify = "entity_id_here"
            >>> attribute_updates = {'property1': new_value1, 'property2': new_value2}
            >>> modified_entity = MyClass.modify(entity_id_to_modify, attribute_updates)
        """
        entity_dict: dict | None = cls.get_response(params={'entity_id': entity_id, 'overwrites': overwrites})

        if entity_dict is None:
            return None

        return cls(**entity_dict)

    @classmethod
    def remove(cls, entity_id: str) -> Self | None:
        """
        Remove an entity by its unique identifier.

        Args:
            entity_id (str): The unique identifier of the entity to remove.

        Returns:
            Self: An instance of the class representing the removed entity.
            None: If not found

        Example:
            To remove an entity by its unique identifier:

            removed_entity = MyClass.remove("entity_id_here")
            >>> if removed_entity:
            >>>     # Handle the removed entity
            >>> else:
            >>>     # Entity not found
        """
        entity_dict: dict | None = cls.get_response(params={'entity_id': entity_id})

        if entity_dict is None:
            return None

        return cls(**entity_dict)

    @classmethod
    def clone(cls, entity_id: str, overwrites: dict) -> Self | None:
        """
        Clone an existing entity by creating a new one based on provided overwrites.

        Args:
            entity_id (str): The unique identifier of the entity to clone.
            overwrites (dict): A dictionary containing attribute updates to apply to the new cloned entity.

        Returns:
            Self: An instance of the class representing the newly cloned entity.
            None: If not found.

        Example:
            To clone an existing entity by creating a new one with attribute overwrites and optional keyword arguments:

            >>> entity_id_to_clone = "entity_id_here"
            >>> attribute_overwrites = {'property1': new_value1, 'property2': new_value2}
            >>> cloned_entity = MyClass.clone(entity_id_to_clone, attribute_overwrites)
        """
        entity_dict: dict | None = cls.get_response(params={'entity_id': entity_id, 'overwrites': overwrites})

        if entity_dict is None:
            return None

        return cls(**entity_dict)

    @classmethod
    def retrieve_many(cls, entity_id_list: list[str], projection: str | list[str] | None = None) -> list[Self | None]:
        """
        Retrieve multiple entities by their unique identifiers.

        Args:
            entity_id_list (List[str]): A list of unique identifiers for the entities to retrieve.
            projection (Union[List, str], optional): A property name as a string or a list of property names.

        Returns:
            List[Self | None]: A list of instances of the class representing the retrieved entities, or None for entities not found.

        Example:
            To retrieve multiple entities by their unique identifiers:

            >>> entity_ids_to_retrieve = ["entity_id1", "entity_id2"]
            >>> retrieved_entities = MyClass.retrieve_many(entity_ids_to_retrieve)
        """
        if not isinstance(entity_id_list, list):
            raise FieldValueError(f"The argument 'entity_id_list' most be a instance of 'list' and not {type(entity_id_list)}.")

        if projection is not None:
            projection = cls._normalize_projection(projection)

        chunks: list[slice] = split_in_slices(len(entity_id_list), settings.ENTITY_RETRIEVE_BATCH_SIZE)
        entities: list[Self | None] = []
        for chunk in chunks:
            entities.extend(cls.inner_retrieve_many(entity_id_list[chunk], projection))

        return entities

    @classmethod
    def inner_retrieve_many(cls, entity_id_list: list[str], projection: str | list[str] | None = None) -> list[Self | None]:
        """
        Retrieve multiple entities by their unique identifiers.

        Args:
            entity_id_list (List[str]): A list of unique identifiers for the entities to retrieve.
            projection (Union[List, str], optional): A property name as a string or a list of property names.

        Returns:
            List[Self | None]: A list of instances of the class representing the retrieved entities, or None for entities not found.
        """

        entities: list[dict | None] = cls.get_response(params={'entity_id_list': entity_id_list, 'projection': projection})

        return [cls(**entity_dict) if entity_dict is not None else None for entity_dict in entities]

    @classmethod
    def create_many(cls, entity_dict_list: list[dict]) -> list[Self | None]:
        """
        Create multiple new entities using provided dictionaries..

        Args:
            entity_dict_list (List[dict]): A list of dictionaries, each representing an entity's attributes.

        Returns:
            List[Self | None]: A list of instances of the class representing the newly created entities, or None for entities not found.

        Example:
            To create multiple entities using a list of dictionaries:

            >>> entity_data_list = [{'property1': value1}, {'property2': value2}]
            >>> created_entities = MyClass.create_many(entity_data_list)
        """
        entities: list[dict | None] = cls.get_response(params={'entity_dict_list': entity_dict_list})

        return [cls(**entity_dict) if entity_dict is not None else None for entity_dict in entities]

    @classmethod
    def modify_many(cls, entity_id_list: list[str], overwrites: dict | list[dict]) -> list[Self | None]:
        """
        Modify multiple existing entities by updating their attributes using the provided overwrites.

        Args:
            entity_id_list (List[str]): A list of unique identifiers for the entities to modify.
            overwrites (Union[dict, List[dict]]): A dictionary or a list of dictionaries containing attribute updates
                to apply to the entities.

        Returns:
            List[Self | None]: A list of instances of the class representing the modified entities, or None for entities not found.

        Example:
            To modify multiple existing entities by updating their attributes with overwrites:

            >>> entity_ids_to_modify = ["entity_id1", "entity_id2"]
            >>> attribute_overwrites = [{'property1': new_value1}, {'property2': new_value2}]
            >>> modified_entities = MyClass.modify_many(entity_ids_to_modify, attribute_overwrites)
        """
        entities: list[dict | None] = cls.get_response(params={'entity_id_list': entity_id_list, 'overwrites': overwrites})

        return [cls(**entity_dict) if entity_dict is not None else None for entity_dict in entities]

    @classmethod
    def remove_many(cls, entity_id_list: list[str]) -> list[str | None]:
        """
        Remove multiple entities by their unique identifiers.

        Args:
            entity_id_list (List[str]): A list of unique identifiers for the entities to remove.

        Returns:
            List[str | None]: A list of unique identifiers for the removed entities, or None for entities not deleted.

        Example:
            To remove multiple entities by their unique identifiers:

            >>> entity_ids_to_remove = ["entity_id1", "entity_id2"]
            >>> MyClass.remove_many(entity_ids_to_remove)
        """
        return cls.get_response(params={'entity_id_list': entity_id_list})

    @classmethod
    def clone_many(cls, entity_id_list: list[str], overwrites: dict | list[dict]) -> list[Self | None]:
        """
        Clone multiple existing entities by creating new ones based on provided overwrites.

        Args:
            entity_id_list (List[str]): A list of unique identifiers for the entities to clone.
            overwrites (Union[dict, List[dict]]): A dictionary or a list of dictionaries containing attribute updates
                to apply to the new cloned entities.

        Returns:
            List[Self | None]: A list of instances of the class representing the newly copied entities, or None for entities not found.

        Example:
            To clone multiple existing entities by creating new ones with attribute overwrites:

            >>> entity_ids_to_clone = ["entity_id1", "entity_id2"]
            >>> attribute_overwrites = [{'property1': new_value1}, {'property2': new_value2}]
            >>> copied_entities = MyClass.clone_many(entity_ids_to_clone, attribute_overwrites)
        """
        entities: list[dict | None] = cls.get_response(params={'entity_id_list': entity_id_list, 'overwrites': overwrites})

        return [cls(**entity_dict) if entity_dict is not None else None for entity_dict in entities]

    def load(self, offset: int = None) -> Self | None:
        """
        Load an entity from the database and return it as an instance of the class.

        Args:
            self (Self): The entity instance to load.
            offset (int, optional): The offset to use for pagination. Defaults to None.

        Returns:
            Self: An instance of the class representing the loaded entity.
            None: If not found.

        Example:
            >>> entity_to_load = MyClass(property1="value1", property2="value2")
            >>> loaded_entity = entity_to_load.load()
        """
        # pylint: disable=no-member
        if self.id:
            return type(self).retrieve(self.id)

        query = self.to_query()
        return query.load(offset=offset)

    def save(self) -> Self:
        """
        Save the entity to the database and return the saved entity as an instance of the class.

        Args:
            self (Self): The entity instance to save.

        Returns:
            Self: An instance of the class representing the saved entity.

        Example:
            To save an entity:

            >>> entity_to_save = MyClass(id="entity_id_here", property1="value1", property2="value2")
            >>> saved_entity = entity_to_save.save()
        """
        entity_dict: dict = self.get_response(self_obj=self)
        return self.__class__(**entity_dict)

    def delete(self) -> Self | None:
        """
        Delete the entity from the database and return the deleted entity as an instance of the class.

        Returns:
            Self: An instance of the class representing the deleted entity.
            None: If not found.

        Example:
            To delete an entity:

            >>> entity_to_delete = MyClass(id="entity_id_here")
            >>> deleted_entity = entity_to_delete.delete()
        """
        entity_dict: dict | None = self.get_response(self_obj=self)

        if entity_dict is None:
            return None

        return self.__class__(**entity_dict)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.unittests import TestCase

from everysk.sdk.engines import __getattr__

###############################################################################
#   Init Test Case Implementation
###############################################################################
# pylint: disable=import-outside-toplevel
# pylint: disable=no-name-in-module
# pylint: disable=unused-import
class EnginesInitTestCase(TestCase):

    def test_get_attribute_imports_expression(self):
        from everysk.sdk.engines.expression import Expression as expected
        from everysk.sdk.engines import Expression
        self.assertEqual(expected, Expression)

    def test_get_attribute_imports_compliance(self):
        from everysk.sdk.engines.compliance import Compliance as expected
        from everysk.sdk.engines import Compliance
        self.assertEqual(expected, Compliance)

    def test_get_attribute_imports_cache(self):
        from everysk.sdk.engines.cache import UserCache as expected
        from everysk.sdk.engines import UserCache
        self.assertEqual(expected, UserCache)

    def test_get_attribute_imports_non_existent_module_raises_attribute_error(self):
        with self.assertRaises(ImportError) as context:
            from everysk.sdk.engines import non_existent_module
        self.assertEqual(str(context.exception), "cannot import name 'non_existent_module' from 'everysk.sdk.engines' (/var/app/src/everysk/sdk/engines/__init__.py)")

    def test_get_attribute_imports_all_modules_path(self):
        from everysk.config import settings
        modules = settings.EVERYSK_SDK_ENGINES_MODULES_PATH.keys()
        self.assertListEqual(list(modules), ['Expression', 'Compliance', 'UserCache', 'MarketData'])



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['BaseEndpoint', 'JSONEndpoint']
from urllib.parse import urlparse
from typing import Any, Generator
import httpx
from starlette.types import Receive, Scope, Send

from everysk.config import settings
from everysk.core.exceptions import HttpError
from everysk.core.log import Logger, LoggerManager, _get_trace_data
from everysk.core.serialize import loads
from everysk.server.requests import Request, JSONRequest
from everysk.server.responses import Response, JSONResponse


log = Logger(__name__)
HTTP_METHODS = ('GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS')
HTTP_STATUS_CODES_LOG = settings.EVERYSK_SERVER_CODES_LOG


class BaseEndpoint:
    # Based in starlette.endpoints.HTTPEndpoint
    ## Private attributes
    _allowed_methods: list[str] = None
    _request_class: Request = Request
    _response_class: Response = Response

    ## Public attributes
    receive: Receive = None
    request: Request = None
    scope: Scope = None
    send: Send = None

    ## Private methods
    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Base class for all endpoints in the application.

        Args:
            scope (Scope): ASGI scope dictionary.
            receive (Receive): ASGI receive data.
            send (Send): ASGI send data.

        Raises:
            HttpError: 500 - Request is not an HTTP request.
        """
        type_request = scope.get('type', '')
        if type_request.lower() != 'http':
            raise HttpError(status_code=500, msg='Request is not an HTTP request.')

        self._allowed_methods = [method for method in HTTP_METHODS if hasattr(self, method.lower())]
        self.receive = receive
        self.request = self._request_class(scope, receive=receive)
        self.scope = scope
        self.send = send

    def __await__(self) -> Generator[Any, None, None]:
        """
        Method to allow the use of the await keyword in the class.
        This method will call the dispatch method and return the result.
        It's the default behavior of the Starlette HTTPEndpoint class.
        Don't change this method.
        """
        return self.dispatch().__await__()

    ## Public sync methods
    def get_http_headers(self) -> dict[str, str]:
        """
        Get the HTTP headers from the request.
        Returns dictionary were the key is the header name in lower case and the value is the header value.
        """
        return dict(self.request.headers)

    def get_http_method_function(self) -> callable:
        """
        Get the function that for the http method of the request.
        If the function doesn't exist, it will return the method_not_allowed function.
        """
        name = self.get_http_method_name()
        return getattr(self, name, self.method_not_allowed)

    def get_http_method_name(self) -> str:
        """
        Get the name of the HTTP method from the request.
        If the request method is HEAD and the class doesn't
        have a head method, it will return get instead.
        """
        if self.request.method == 'HEAD' and not hasattr(self, 'head'):
            name = 'get'
        else:
            name = self.request.method.lower()

        return name

    ## Public async methods
    async def dispatch(self) -> None:
        """
        Main method that will always be executed for each request, takes
        the function related to the HTTP method of the request and executes it.
        """
        # Because the ASGI protocol copy the context to the event loop
        # for every request, we create an empty LoggerManager to avoid
        # shared values between requests.
        with LoggerManager(http_headers={}, http_payload={}, labels={}, stacklevel=None, traceback=''):
            headers = self.get_http_headers()
            # Insert the headers in the Logger Context to propagate them to the logs
            with LoggerManager(http_headers=headers):
                try:
                    response = await self.get_http_response()
                except Exception as error: # pylint: disable=broad-except
                    # If something goes wrong, we catch the exception and return a response
                    response = await self.get_http_exception_response(error)

                    # We only log internal server errors in GCP
                    if getattr(error, 'status_code', 500) in HTTP_STATUS_CODES_LOG:
                        payload = await self.get_http_payload()
                        # Headers are already in the LoggerManager
                        log.error(str(error), extra={'http_payload': payload})

                await response(self.scope, self.receive, self.send)

        # To avoid shared values between requests, we reset the LoggerManager
        LoggerManager.reset()

    async def get_http_exception_response(self, error: Exception) -> Response:
        """
        Method to return a response when an exception is raised during the request.

        Args:
            error (Exception): The exception raised during the request.
        """
        status_code = getattr(error, 'status_code', 500)
        return self._response_class(str(error), status_code=status_code)

    async def get_http_payload(self) -> bytes:
        """
        Get the HTTP payload from the request.
        The payload is the body of the request and it's a bytes object.
        """
        return await self.request.body()

    async def get_http_response(self) -> Response:
        """
        Get the correct function for the HTTP method of the request
        and execute it to create a response.
        If the method doesn't exist, it will return a 405 response.
        """
        method_function = self.get_http_method_function()
        response = await method_function()

        if not isinstance(response, Response):
            response = self._response_class(response)

        return response

    async def method_not_allowed(self) -> None:
        """
        Default method for when the HTTP method is not found in the class.

        Raises:
            HttpError: 405 - Method not allowed
        """
        raise HttpError(status_code=405, msg=f'Method {self.request.method} not allowed.')


class JSONEndpoint(BaseEndpoint):
    ## Private attributes
    _request_class: JSONRequest = JSONRequest
    _response_class: JSONResponse = JSONResponse

    ## Public attributes
    rest_key_name: str = Undefined
    rest_key_value: str = Undefined

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Class to handle JSON requests and responses.
        Inherit from this class and implement the HTTP methods to create an endpoint.

        Args:
            scope (Scope): ASGI scope dictionary.
            receive (Receive): ASGI receive data.
            send (Send): ASGI send data.
        """
        super().__init__(scope, receive, send)

        if self.rest_key_name is Undefined:
            self.rest_key_name = settings.EVERYSK_SERVER_REST_KEY_NAME

        if self.rest_key_value is Undefined:
            self.rest_key_value = settings.EVERYSK_SERVER_REST_KEY_VALUE

    def check_rest_key(self) -> bool:
        """
        Check if the rest key is present in the request headers and if it's the correct value.
        If the rest key name or value is not set, it will always return True.
        """
        if not self.rest_key_name or not self.rest_key_value:
            return True

        rest_key_value = self.request.headers.get(self.rest_key_name)
        return rest_key_value == self.rest_key_value

    async def get_http_exception_response(self, error: Exception) -> JSONResponse:
        """
        Method to return a JSONResponse when an exception is raised during the request.
        The trace_id is added to the response to help with debugging.

        Args:
            error (Exception): The exception raised during the request.

        Returns:
            JSONResponse: A JSONResponse with the error message, status code and trace_id.
        """
        trace_data = _get_trace_data(headers=self.get_http_headers())
        msg = str(error)
        status_code = getattr(error, 'status_code', 500)
        return self._response_class(
            {'error': msg, 'code': status_code, 'trace_id': trace_data['trace_id']},
            status_code=status_code
        )

    async def get_http_payload(self) -> Any:
        """
        Get the HTTP payload from the request and deserialize it to a
        Python object or an empty dict if the request.body is empty.
        """
        body = await super().get_http_payload()
        if body:
            return loads(body, protocol='json', use_undefined=True)

        return {}

    async def get_http_response(self) -> JSONResponse:
        """
        Changes the return of the get_http_response method to return a JSONResponse.
        If the response is not a Response object, it will be converted to a JSONResponse
        otherwise it will be returned as is.
        If the rest key is incorrect, it will raise a 401 error.

        Raises:
            HttpError: 401 - Unauthorized access to this resource.
        """
        if not self.check_rest_key():
            raise HttpError(status_code=401, msg='Unauthorized access to this resource.')

        response = await super().get_http_response()
        return response


class HealthCheckEndpoint(JSONEndpoint):
    """
    Endpoint to check if the service is running.
    By default, it will return a JSONResponse with the status 'SENTA_A_PUA'.
    """
    default_response: dict = {'status': 'SENTA_A_PUA'}
    # These are set to None so the endpoint can be accessed without the rest key
    rest_key_name: str = None
    rest_key_value: str = None

    async def get(self) -> JSONResponse:
        return JSONResponse(self.default_response)

    async def post(self) -> JSONResponse:
        return JSONResponse(self.default_response)


class RedirectEndpoint(BaseEndpoint):
    """
    Endpoint to redirect requests to another host.
    We use this endpoint to redirect requests to another host
    and return the response to the client, acting as a proxy.

    Raises:
        ValueError: If the host_url is not set in the class.
    """
    host_url: str = settings.EVERYSK_SERVER_REDIRECT_URL
    timeout: int = 600

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope, receive, send)
        if not self.host_url:
            raise ValueError('host_url is required for redirect endpoint.')

    def get_full_url(self):
        """
        Get the full URL to be used in the connection.
        This method will return the URL with the host, path and query string.
        """
        url = self.request.url
        result = f'{self.host_url}{url.path}'
        if url.query:
            result = f'{result}?{url.query}'
        return result

    def get_host(self):
        """
        Get the host from the host_url attribute.
        """
        url = urlparse(self.host_url)
        return url.netloc

    def get_request_headers(self):
        """
        Get the headers received in the request and update the Host header with the destination host.
        """
        headers = dict(self.request.headers)

        # We need to update the Host header with the destination host
        headers['host'] = self.get_host()

        return headers

    def get_response_headers(self, response: httpx.Response):
        """
        Get the headers from the redirected response and keep only the content_type to be used in the response.

        Args:
            response (httpx.Response): The response from the redirect request.
        """
        return {'content-type': response.headers.get('content-type')}

    def get_timeout(self) -> httpx.Timeout:
        """
        Return the timeout to be used in the connection.
        """
        timeout = httpx.Timeout(
            timeout=30, # Default timeout for all operations
            read=self.timeout # Timeout for reading the response
        )
        return timeout

    def make_response(self, response: httpx.Response):
        """
        Create a Response or JSONResponse object based on the content type of the response.

        Args:
            response (httpx.Response): The response from the redirect request.
        """
        content_type = response.headers.get('content-type')
        cls = Response if 'json' in content_type else JSONResponse
        headers = self.get_response_headers(response=response)

        return cls(status_code=response.status_code, content=response.content, headers=headers)

    async def get(self):
        """
        HTTP GET method to redirect the request to another host.
        """
        with httpx.Client(headers=self.get_request_headers()) as connection:
            response = connection.get(self.get_full_url(), timeout=self.get_timeout())

        return self.make_response(response=response)

    async def post(self):
        """
        HTTP POST method to redirect the request to another host.
        """
        body = await self.request.body()
        with httpx.Client(headers=self.get_request_headers()) as connection:
            response = connection.post(self.get_full_url(), content=body, timeout=self.get_timeout())

        return self.make_response(response=response)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.api_resource import (
    DeletableAPIResource,
    RetrievableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource
)

###############################################################################
#   Report Template Implementation
###############################################################################
class ReportTemplate(
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    UpdateableAPIResource
):
    """
    Represents a report template.

    This class inherits from the following API resource classes:
    - RetrievableAPIResource
    - ListableAPIResource
    - DeletableAPIResource
    - UpdateableAPIResource
    """
    pass



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any

from everysk.config import settings
from everysk.sdk.base import BaseSDK

###############################################################################
#   Cache Implementation
###############################################################################
class UserCache(BaseSDK):

    def get(self, key: str, prefix: str = '') -> Any:
        """
        Get the value of a key from the cache.

        Args:
            key (str): The key to get the value of.
            prefix (str): The prefix of the key.

        Returns:
            Any: The value of the key.

        Example:
            >>> cache = UserCache()
            >>> cache.get('key')
        """
        return self.get_response(self_obj=self, params={'key': key, 'prefix': prefix})

    def get_multi(self, keys: list[str], prefix: str = '') -> dict[str, Any]:
        """
        Get the values of multiple keys from the cache.

        Args:
            keys (list[str]): The keys to get the values of.
            prefix (str): The prefix of the keys.

        Returns:
            dict[str, Any]: The values of the keys.

        Example:
            >>> cache = UserCache()
            >>> cache.get_multi(['key1', 'key2'])
            {
                'key1': 'value1',
                'key2': 'value2'
            }
        """
        return self.get_response(self_obj=self, params={'keys': keys, 'prefix': prefix})

    def set(self, key: str, data: Any, time: int = settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME, prefix: str = '') -> bool:
        """
        Set the value of a key in the cache.

        Args:
            key (str): The key to set the value of.
            data (Any): The value to set.
            time (int): The expiration time of the key in seconds.
            prefix (str): The prefix of the key.

        Returns:
            bool: True if the key is set, False otherwise.

        Example:
            >>> cache = UserCache()
            >>> cache.set('key', 'value')
        """
        if not isinstance(time, int) or time <= 0 or time > settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME:
            raise ValueError('Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.')
        return self.get_response(self_obj=self, params={'key': key, 'data': data, 'time': time, 'prefix': prefix})

    def set_multi(self, data_dict: dict, time: int = settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME, prefix: str = '') -> list:
        """
        Set the values of multiple keys in the cache.

        Args:
            data_dict (dict): The keys and values to set.
            time (int): The expiration time of the keys in seconds.
            prefix (str): The prefix of the keys.

        Returns:
            list: The keys that are set.

        Example:
            >>> cache = UserCache()
            >>> cache.set_multi({'key1': 'value1', 'key2': 'value2'})
            ['key1', 'key2']
        """
        if not isinstance(time, int) or time <= 0 or time > settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME:
            raise ValueError('Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.')
        return self.get_response(self_obj=self, params={'data_dict': data_dict, 'time': time, 'prefix': prefix})

    def incr(self, key: str, delta: int = 1, initial_value: Any = None, time: int = settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME) -> int:
        """
        Increment the value of a key in the cache.

        Args:
            key (str): The key to increment the value of.
            delta (int): The amount to increment the value by.
            initial_value (Any): The initial value of the key.
            time (int): The expiration time of the key in seconds.

        Returns:
            int: The new value of the key.

        Example:
            >>> cache = UserCache()
            >>> cache.incr('key', 1, 0)
            1
        """
        if not isinstance(time, int) or time <= 0 or time > settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME:
            raise ValueError('Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.')
        return self.get_response(self_obj=self, params={'key': key, 'delta': delta, 'initial_value': initial_value, 'time': time})

    def decr(self, key: str, delta: int = 1, initial_value: Any = None, time: int = settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME) -> int | None:
        """
        Decrement the value of a key in the cache.

        Args:
            key (str): The key to decrement the value of.
            delta (int): The amount to decrement the value by.
            initial_value (Any): The initial value of the key.
            time (int): The expiration time of the key in seconds.

        Returns:
            int: The new value of the key.

        Example:
            >>> cache = UserCache()
            >>> cache.decr('key', 1, 0)
            0
        """
        if not isinstance(time, int) or time <= 0 or time > settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME:
            raise ValueError('Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.')
        return self.get_response(self_obj=self, params={'key': key, 'delta': delta, 'initial_value': initial_value, 'time': time})

    def delete(self, key: str) -> bool:
        """
        Delete a key from the cache.

        Args:
            key (str): The key to delete.

        Returns:
            bool: True if the key is deleted, False otherwise.

        Example:
            >>> cache = UserCache()
            >>> cache.delete('key')
            True
        """
        return self.get_response(self_obj=self, params={'key': key})

    def delete_multi(self, keys: list[str]) -> bool:
        """
        Delete multiple keys from the cache.

        Args:
            keys (list[str]): The keys to delete.

        Returns:
            bool: True if the keys are deleted, False otherwise.

        Example:
            >>> cache = UserCache()
            >>> cache.delete_multi(['key1', 'key2'])
            True
        """
        return self.get_response(self_obj=self, params={'keys': keys})



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from datetime import date, datetime # pylint: disable=import-self
from typing import Any, Union, Self

from everysk.core.datetime import date_settings
from everysk.core.datetime.date_expression import DateExpressionMixin
from everysk.core.datetime.date_mixin import DateMixin


class Date(DateMixin, DateExpressionMixin, date):

    @classmethod
    def ensure(cls, value: Any) -> Self:
        """
        Ensure that a value is a Date object or convert it to one.

        This class method takes a value and ensures that it is a Date object. If the value is a datetime.date object, it will be
        converted to a Date object by extracting the year, month, and day components. If the value is already a Date object, it
        will be returned as is. If the value is neither a Date object nor a datetime.date object, a ValueError is raised.

        Args:
            value (Any): The value to ensure as a Date object.

        Raises:
            ValueError: If the value cannot be converted to a Date object.

        Returns:
            date: A Date object representing the same date as the input value.

        Criteria:
            - The method accepts a value that is either a Date object or a datetime.date object.
            - It ensures that the value is a Date object or converts it to one if necessary.
            - It raises a ValueError if the value cannot be converted to a Date object.

        Example:
            Ensure a value is a Date object:
            >>> date_value = Date(2023, 7, 31)
            >>> Date.ensure(date_value)
            Date(2023, 7, 31)

            >>> datetime_date = datetime.date(2023, 7, 31)
            >>> Date.ensure(datetime_date)
            Date(2023, 7, 31)

        Example of exception message:
            ValueError: Invalid instantiation of class 'Date' from 'str'
        """
        if value.__class__ != cls:
            try:
                value = cls(value.year, value.month, value.day)
            except Exception:
                raise ValueError(f"Invalid instantiation of class '{cls.__name__}' from '{value.__class__.__name__}'") # pylint: disable=raise-missing-from

        return value

    @classmethod
    def fromisoformat(cls, date_string: str) -> Self:
        """
        Convert an ISO formatted date string to a Date object.

        This class method takes an ISO formatted date string and converts it to a Date object.

        Args:
            date_string (str):
                The ISO formatted date string to convert to a Date object.

        Raises:
            ValueError: If the input date string does not have the expected format.

        Returns:
            date:
                A Date object representing the date extracted from the input ISO formatted date string.

        Criteria:
            - The method accepts an ISO formatted date string.
            - It converts the date string to a Date object.
            - It raises a ValueError if the input date string does not have the expected format.

        Example:
            Convert an ISO formatted date string to a Date object:
            >>> input_string = '20230101'
            >>> Date.fromisoformat(input_string)
            Date(2023, 1, 1)

            >>> input_string = '2023-01-01'
            >>> Date.fromisoformat(input_string)
            Date(2023, 1, 1)

        Example of exception message:
            ValueError: Invalid isoformat string: '2023-01-01T00:00:00'.
        """
        if date_string and len(date_string) == 8:
            return cls.strptime(date_string)

        return super().fromisoformat(date_string)

    # antes retornava string, agora retorna self date
    # agora start_date é date
    @classmethod
    def fromordinal(cls, date_ordinal: int, start_date: 'Date' = None) -> Self:
        """
        Convert the ordinal day count since a base date to a Date object.

        This static method takes an `date_ordinal`, which is the number of days since a specified `start_date`,
        and converts it into a Date object.

        Args:
            date_ordinal (int):
                The ordinal day count since the base date specified by `start_date`.
            start_date (Date, optional):
                The base date to start counting from. Defaults to January 1, 1 AD (00010101).

        Raises:
            TypeError:
                If the input date_ordinal is not an integer.
            AttributeError:
                If the start date is not a Date object or None.

        Returns:
            Date:
                A Date object representing the date calculated from the ordinal day count and start_date.

        Criteria:
            - The method accepts an ordinal day count and an optional start_date.
            - It converts the ordinal day count to a Date object.
            - If start_date is provided, it uses it as the base date for the calculation.

        Example:
            Convert an ordinal day count to a Date object:
            >>> Date.fromordinal(1)
            Date(1, 1, 1)

            >>> Date.fromordinal(365)
            Date(1, 12, 31)

        Example of exception message:
            TypeError: can only concatenate str (not "int") to str
        """
        if start_date:
            date_ordinal += start_date.toordinal() - 1

        return super().fromordinal(date_ordinal)

    @classmethod
    def strptime(cls, date_string: str, format: str = date_settings.DEFAULT_DATE_FORMAT) -> Self:
        """
        Convert a string to a Date object using the specified format.

        This static method takes a string and attempts to convert it into a Date object using the provided
        format. If the conversion is successful, the resulting Date object is returned. If the input string is
        empty or cannot be converted, a ValueError is raised.

        Args:
            date_string (str):
                The input string to be converted to a Date object.
            format (str, optional):
                The format string used to interpret the input string. Defaults to DEFAULT_DATE_FORMAT ('%Y%m%d').

        Raises:
            ValueError: If the input string is empty or cannot be converted to a Date object using the specified format.

        Returns:
            Date:
                A Date object representing the converted date if successful.

        Example:
            >>> date_str = '20230731'
            >>> Date.string_to_date(date_str)
            Date(2023, 7, 31)

        Example of exception message:
            'ValueError: time data 'invalid-date' does not match format '%Y%m%d''
        """
        datetime_ = datetime.strptime(date_string, format)
        return cls(datetime_.year, datetime_.month, datetime_.day)

    @classmethod
    def strptime_or_null(cls, date_string: str, format: str = date_settings.DEFAULT_DATE_FORMAT) -> Union[Self, None]:
        """
        Convert a string to a Date object using the specified format or return None.

        This class method takes a string and attempts to convert it into a Date object using the provided
        format. If the conversion is successful, the resulting Date object is returned. If the input string is
        empty or cannot be converted, the method returns None.

        Args:
            date_string (str):
                The input string to be converted to a Date object.
            format (str, optional):
                The format string used to interpret the input string. Defaults to DEFAULT_DATE_FORMAT ('%Y%m%d').

        Returns:
            Date or None:
                A Date object representing the converted date if successful, or None if the input string is empty or
                cannot be converted.

        Example:
            To convert a string to a Date object:
            >>> date_string = '20230731'
            >>> date_obj = Date.strptime_or_null(date_string)
        """
        date_ = None

        try:
            date_ = Date.strptime(date_string, format=format)
        except Exception: # pylint: disable=broad-exception-caught
            pass

        return date_

    @classmethod
    def is_string_date(cls, date_string: str, format: str = date_settings.DEFAULT_DATE_FORMAT) -> bool:
        """
        Check if a string is a valid representation of a date.

        This class method checks if the input string is a valid representation of a date
        using the specified date format. If the input string can be successfully converted to a
        Date object using the provided format, the method returns True. Otherwise, it returns False.

        Args:
            date_string (str):
                The string to check for validity as a date representation.
            format (str, optional):
                The format string used to interpret the input string. Defaults to DEFAULT_DATE_FORMAT ('%Y%m%d').

        Returns:
            bool:
                True if the input string is a valid date representation, False otherwise.

        Example:
            To check if a string is a valid date representation:
            >>> date_string = '20230731'
            Date.is_string_date(date_string)
            True

            >>> date_string = '2023-05-01'
            Date.is_string_date(date_string)
            False

            >>> date_string = '2023-05-01'
            Date.is_string_date(date_string, format='%Y-%m-%d')
            True
        """
        return bool(cls.strptime_or_null(date_string, format=format))

    def strftime(self, format: str = date_settings.DEFAULT_DATE_FORMAT) -> str:
        """
        Convert a `Date` object to a date string representation.

        This instance method takes a `Date` object and converts it to a date string representation
        using the specified date `format`. If no `format` is provided, the method will use the Everysk's default date
        format `DEFAULT_DATE_FORMAT`.

        Args:
            format (str, optional): The desired date format. This should be a string representing the desired format
                                using the appropriate format codes. (default is `DEFAULT_DATE_FORMAT`)

        Raises:
            TypeError: If a invalid argument is provided.

        Returns:
            str: The date string representation of the input `Date` object in the specified date format.

        Example:
            To convert a `Date` object to a date string:
            >>> date_obj = Date(2023, 7, 31)
            >>> date_obj.strftime(format='%d/%m/%Y')
            '31/07/2023'

        Example of exception message:
            TypeError: strftime() argument 1 must be str, not int
        """
        return super().strftime(format)

    @classmethod
    def strftime_or_null(cls, date_: Any, format: str = date_settings.DEFAULT_DATE_FORMAT) -> Union[str, None]:
        """
        Convert a Date object to a string representation or return None.

        This class method takes a Date object and converts it to a string representation
        using the default date format ('%Y%m%d'), which represents dates in the format 'YYYYMMDD'.
        If the input date is None, the method returns None.

        Args:
            date_ (Date or None): The Date object to convert to a string representation, or None.
            format (str, optional): The desired date format. This should be a string representing the desired format
                                using the appropriate format codes. (default is `DEFAULT_DATE_FORMAT`)

        Returns:
            str or None: A string representation of the input date in the specified date format, or None if the input date is None.

        Example:
            To convert a Date object to a date string or return None:
            >>> import datetime
            >>> date_obj = Date(2023, 7, 31)
            >>> Date.strftime_or_null(date_obj)
            '20230731'

            >>> Date.strftime_or_null(None)
            None
        """
        if date_ is not None:
            try:
                date_ = date_.strftime(format=format)
            except Exception:
                date_ = None

        return date_

    @classmethod
    def is_date(cls, value: Any) -> bool:
        """
        Check if a value is a Date object.

        This class method checks if the input value is an instance of the `Date` class.

        Args:
            value (Any): The value to be checked.

        Returns:
            bool: True if the input value is a Date object, False otherwise.

        Example:
            >>> date_obj = Date(2023, 7, 31)
            >>> Date.is_date(date_obj)
            True

            >>> date_time_obj = DateTime(2023, 7, 31, 12, 0)
            >>> Date.is_date(date_time_obj)
            False
        """
        if value is not None:
            if value.__class__ in (Date, date):
                return True

        return False

    def is_today(self) -> bool:
        """
        Check if the current date corresponds to the date represented by this Date object.

        This method compares the date represented by this Date object with the current date and returns True if they are the same.

        Returns:
            bool: True if this Date object represents the current date, False otherwise.

        Example:
            >>> today = Date.today()
            >>> today.is_today()
            True

            >>> future_date = Date(2023, 12, 31)
            >>> future_date.is_today()
            False
        """
        return self == Date.today()

    def date(self) -> Self:
        """
        Return Date object with same year, month and day. Used to maintain consistency between DateTime and Date instances

        Returns:
            date: A Date object representing the same date as the input value.

        Example:
            >>> today = Date(2023, 12, 31)
            >>> today.date()
            Date(2023, 12, 31)
        """
        return self

    ####################################
    ### DATE RANGE
    ####################################

    @classmethod
    def date_range(cls, range_type: str, period: str, n_period: int, start_date: 'Date', end_date: 'Date') -> tuple:
        """
        Get the start date and end date based on specified parameters.

        This method returns a tuple of two dates (start_date and end_date) based on the provided parameters. It is useful for generating date ranges based on different requirements.

        Args:
            range_type (str): Specifies the type of range needed. Possible values: 'single_date', 'custom_range', 'n_period', 'period_to_date', 'end_period_to_date'.
            period (str): Specifies the type of period in case `range_type` is 'n_period' or 'period_to_date'. Possible values: 'days', 'weeks', 'months', 'years', 'mtd', 'qtd', 'ytd'.
            n_period (int): Specifies the number of periods to be considered in case `range_type` is 'n_period'.
            start_date (Date): The start date of the custom range in case `range_type` is 'custom_range'.
            end_date (Date): The end date of the custom range in case `range_type` is 'custom_range' or 'single_date'.

        Returns:
            tuple of str: A tuple of start date and end date in string format ('YYYYMMDD').

        Raises:
            ValueError: If an invalid `endDate`, `startDate`, `nPeriod`, `period`, or `range_type` is provided.

        Example:
            >>> Date.date_range('single_date', None, None, None, Date(2023, 1, 10))
            ('20230110', '20230110')

            >>> Date.date_range('custom_range', None, None, Date(2023, 1, 1), Date(2023, 1, 10))
            ('20230101', '20230110')

            >>> Date.date_range('n_period', 'months', 3, None, Date(2023, 1, 10))
            ('20231010', '20230110')

            >>> Date.date_range('period_to_date', 'ytd', None, None, Date(2023, 12, 31))
            ('20230101', '20231231')
        """
        if start_date and start_date.__class__ != Date:
            raise ValueError('Invalid start_date type')

        if end_date and end_date.__class__ != Date:
            raise ValueError('Invalid end_date type')

        if not end_date:
            end_date = Date.today()

        date_ = end_date

        if range_type == 'single_date':
            start_date = end_date

        elif range_type == 'custom_range':
            if not start_date:
                start_date = cls.ensure(date_settings.MARKET_START_DATE)

        elif range_type == 'n_period':
            delta = -int(n_period)

            if period == 'days':
                start_date = date_.delta(period=delta, periodicity='D')
            elif period == 'weeks':
                start_date = date_.delta(period=delta, periodicity='W')
            elif period == 'months':
                start_date = date_.delta(period=delta, periodicity='M')
            elif period == 'years':
                start_date = date_.delta(period=delta, periodicity='Y')
            else:
                raise ValueError('invalid n period')

        elif range_type in ('period_to_date', 'end_period_to_date'):
            if period == 'mtd':
                start_date = end_date
                start_date = start_date.replace(day=1)
            elif period == 'qtd':
                start_date = Date(date_.year, 3 * date_.quarter - 2, 1)
            elif period == 'ytd':
                start_date = end_date
                start_date = start_date.replace(month=1, day=1)
            else:
                raise ValueError('invalid period to date')

            if range_type == 'end_period_to_date':
                start_date = start_date.delta(period=-1, periodicity='B')
        else:
            raise ValueError('invalid range type')

        assert (start_date and end_date)

        return (start_date, end_date)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.object import BaseDict

###############################################################################
#   Worker Base Implementation
###############################################################################
class WorkerBase(BaseDict):

    workspace: str = None
    script_inputs: BaseDict = None
    inputs_info: BaseDict = None
    worker_type: str = None
    parallel_info: BaseDict = None
    worker_id: str = None
    workflow_id: str = None
    workflow_execution_id: str = None

    def __init__(self, args: BaseDict = Undefined, **kwargs):
        if args is Undefined:
            args = {}
        kwargs.update(args)
        super().__init__(**kwargs)

    def handle_inputs(self) -> None:
        if self.script_inputs:
            for key, value in self.script_inputs.items():
                setattr(self, key, value)

    def handle_tasks(self) -> None:
        raise NotImplementedError()

    def handle_outputs(self) -> BaseDict:
        raise NotImplementedError()

    def run(self) -> BaseDict:
        self.handle_inputs()
        self.handle_tasks()
        return self.handle_outputs()



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime
from everysk.core.exceptions import RequiredError, SDKValueError
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.file.base import File


###############################################################################
#   File TestCase Implementation
###############################################################################
class FileTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'file_1234567891011211234567890',
            'name': 'SampleFile',
            'tags': ['tag1', 'tag2'],
            'description': 'Description',
            'link_uid': 'link_uid',
            'workspace': 'main',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'data': 'base64data',
            'url': '/1234567891011211234567890',
            'content_type': 'text/csv',
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'hash': '4a0ed08522000734bb845f5d72673d9d113c8236',
        }
        self.file = File(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(File.get_id_prefix(), settings.FILE_ID_PREFIX)

    def test_validate(self):
        expected_data = compress({'class_name': 'File', 'method_name': 'validate', 'self_obj': self.file.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        file: File = self.file.copy()

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            file.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_validate_transient(self):
        expected_data = compress({'class_name': 'File', 'method_name': 'validate_transient', 'self_obj': None, 'params': {'entity_dict': self.file.to_dict(add_class_path=True)}}, protocol='gzip', serialize='json')
        file: File = self.file.copy()

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            File.validate_transient(file)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_entity_to_query(self):
        file: File = self.file.copy()
        self.assertRaisesRegex(
            SDKValueError,
            "Can't filter by Name and Tags at the same time",
            file.to_query
        )

        # Test with name
        file.pop('tags')
        file.pop('link_uid')
        file.pop('url')
        query = file.to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('workspace', '=', 'main'),
            ('name', '<', 'samplefilf'),
            ('name', '>=', 'samplefile'),
            ('date', '>=', self.sample_data['date'].replace(hour=0, minute=0, second=0, microsecond=0)),
            ('date', '<=', self.sample_data['date'].replace(hour=23, minute=59, second=59, microsecond=999999)),
        ])
        self.assertEqual(query.order, [])
        self.assertEqual(query.projection, None)
        self.assertEqual(query.limit, None)
        self.assertEqual(query.offset, None)
        self.assertEqual(query.page_size, None)
        self.assertEqual(query.page_token, None)
        self.assertEqual(query.distinct_on, [])

        # Test with tags
        file.update({'name': None, 'tags': ['tag1', 'tag2']})
        query = file.to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('workspace', '=', 'main'),
            ('date', '>=', self.sample_data['date'].replace(hour=0, minute=0, second=0, microsecond=0)),
            ('date', '<=', self.sample_data['date'].replace(hour=23, minute=59, second=59, microsecond=999999)),
            ('tags', '=', 'tag1'),
            ('tags', '=', 'tag2')
        ])

        # Test with link_uid
        file.update({'tags': None, 'link_uid': 'link_uid'})
        query = file.to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('workspace', '=', 'main'),
            ('link_uid', '=', 'link_uid'),
             ('date', '>=', self.sample_data['date'].replace(hour=0, minute=0, second=0, microsecond=0)),
            ('date', '<=', self.sample_data['date'].replace(hour=23, minute=59, second=59, microsecond=999999))
        ])

        # Test with url
        file.update({'link_uid': None, 'url': '/1234567891011211234567890'})
        query = file.to_query() #pylint: disable=protected-access
        self.assertListEqual(query.filters, [
            ('workspace', '=', 'main'),
              ('date', '>=', self.sample_data['date'].replace(hour=0, minute=0, second=0, microsecond=0)),
            ('date', '<=', self.sample_data['date'].replace(hour=23, minute=59, second=59, microsecond=999999)),
            ('url', '=', '/1234567891011211234567890')
        ])

    def test_generate_url(self):
        file: File = self.file.copy()
        url = file.generate_url()
        with mock.patch('everysk.sdk.entities.file.base.generate_unique_id') as mock_generate_unique_id:
            mock_generate_unique_id.return_value = '1234567891011211234567890'
            url = file.generate_url()
        self.assertEqual(url, '/1234567891011211234567890')

    def test_required_fields(self):
        file: File = self.file.copy()
        file.id = None
        self.assertRaisesRegex(
            RequiredError,
            'The id attribute is required.',
            file.validate_required_fields
        )

        file.id = 'file_1234567891011211234567890'
        file.name = None
        self.assertRaisesRegex(
            RequiredError,
            'The name attribute is required.',
            file.validate_required_fields
        )

        file.name = 'SampleFile'
        file.workspace = None
        self.assertRaisesRegex(
            RequiredError,
            'The workspace attribute is required.',
            file.validate_required_fields
        )

        file.workspace = 'main'
        file.date = None
        self.assertRaisesRegex(
            RequiredError,
            'The date attribute is required.',
            file.validate_required_fields
        )

        file.date = DateTime(2023, 9, 9, 9, 9, 9, 9)
        file.hash = None
        self.assertRaisesRegex(
            RequiredError,
            'The hash attribute is required.',
            file.validate_required_fields
        )

        file.hash = '4a0ed08522000734bb845f5d72673d9d113c8236'
        file.data = None
        self.assertRaisesRegex(
            RequiredError,
            'The data attribute is required.',
            file.validate_required_fields
        )

        file.data = 'base64data'
        file.content_type = None
        self.assertRaisesRegex(
            RequiredError,
            'The content_type attribute is required.',
            file.validate_required_fields
        )

        file.content_type = 'text/csv'
        file.url = None
        self.assertRaisesRegex(
            RequiredError,
            'The url attribute is required.',
            file.validate_required_fields
        )

    def test_check_query_raises_sdk_error(self):
        with self.assertRaises(SDKValueError) as context:
            file: File = self.file.copy()
            file['tags'] = []
            file['link_uid'] = ''
            file.to_query(limit=1)

        self.assertEqual(str(context.exception), "Can't filter by URL and Name, Tags or Link UID at the same time")

    def test_to_dict_file(self):
        result = self.file.to_dict()
        expected_data = {
            'id': 'file_1234567891011211234567890',
            'name': 'SampleFile',
            'tags': ['tag1',
            'tag2'],
            'description': 'Description',
            'link_uid': 'link_uid',
            'workspace': 'main',
            'date': '20230909',
            'data': 'base64data',
            'url': '/file/1234567891011211234567890',
            'content_type': 'text/csv',
            'version': '1.0',
            'hash': '4a0ed08522000734bb845f5d72673d9d113c8236',
            'date_time': '20230909 09:09:09',
            'created': 1694250549,
            'updated': 1694250549
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(result, expected_data)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.api_resource import (
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    CreateableAPIResource,
    UpdateableAPIResource,
    FilterableAPIResource
)
from everysk.api import utils

###############################################################################
#   Datastore Implementation
###############################################################################
class Datastore(
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    CreateableAPIResource,
    UpdateableAPIResource,
    FilterableAPIResource
):
    @classmethod
    def explore(cls, **kwargs):
        """
        Explores the Datastore resource by executing a query with specified parameters to retrieve or manipulate data
        within a given workspace.

        Args:
            user_id (int):
                The ID of the user initiating the request, ensuring that the operation is conducted within
                the user's context and permissions.

            user_role (str):
                The role of the user, which may define the access level and capabilities within the explore operation.

            user_time_zone (str):
                The user's time zone, which can influence how time-based data is retrieved or displayed.

            model (object):
                The model object or identifier specifying the data or views to be explored within the datastore.
                This parameter is mandatory and should start with 'view_' if it's a string object.

            overwrites (object, optional):
                An optional object containing properties to overwrite in the model. This allows
                for dynamic modifications to the model during the exploration.

            workspace (str):
                The workspace identifier within which the exploration is to be conducted. This parameter is mandatory
                and specifies the context or environment for the operation.

        Returns:
            dict: The response from the API request.

        Example:
            >>> response = ClassName.explore(
            >>> ... user_id=12345,
            >>> ... user_role='analyst',
            >>> ... user_time_zone='UTC',
            >>> ... model={'some_model_key': 'some_model_value'},
            >>> ... overwrites={'key_to_overwrite': 'new_value'},
            >>> ... workspace='my_workspace'
            >>> )
            >>> print(response)
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'/{cls.class_name_list()}/explore'
        response = api_req.post(url, kwargs)
        return response
